﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Problems.DataAnalysis.Symbolic;

namespace HeuristicLab.EvolutionaryTracking {

  internal static class Util {
    public static string Label(this ISymbolicExpressionTreeNode node) {
      var v = node as VariableTreeNode;
      return v == null ? node.Symbol.Name : v.VariableName;
    }

    public static string ToPrefixString(this ISymbolicExpressionTreeNode node) {
      if (node.SubtreeCount == 0) return node.Label() + " ";
      var sb = new StringBuilder();

      sb.Append(node.Label().Substring(0, 3) + " ( ");
      foreach (var s in node.Subtrees) {
        sb.Append(s.ToPrefixString());
      }
      sb.Append(") ");
      return sb.ToString();
    }

    public static void WriteDot(this IGenericGraph<SymbolNode> graph, string path) {
      const double minNorm = 1.0, maxNorm = 3.0;

      var arcs = graph.Nodes.Where(node => node.OutEdges != null).SelectMany(node => node.OutEdges).ToList();
      double totalArcWeight = arcs.Sum(arc => arc.Weight);
      double totalNodeWeight = graph.Nodes.Sum(node => node.Weight);
      var min = arcs.Min(x => x.Weight);
      var max = arcs.Max(x => x.Weight);

      var normalizedWeights = arcs.Select(arc => new KeyValuePair<SymbolArc, double>(arc, (arc.Weight - min) * (maxNorm - minNorm) / (max - min) + minNorm))
                                  .ToDictionary(x => x.Key, x => x.Value);

      using (var file = new StreamWriter(path)) {
        const string shape = "rectangle";

        file.WriteLine("digraph {");

        var sb = new StringBuilder();
        foreach (var node in graph.Nodes) {
          double relativeNodeFrequency = node.Weight / totalNodeWeight;
          sb.Clear();
          sb.Append(node.Label + " (" + String.Format("{0:0.00}", node.AverageArity) + ")\\n" + String.Format("{0:0.00%}", relativeNodeFrequency));
          if (node.Positions != null) {
            double posSum = node.Positions.Sum(x => x.Value);
            sb.Append("\\n");
            foreach (var pos in node.Positions)
              sb.Append(String.Format("{0:00%} ", pos.Value / posSum));
          }

          file.WriteLine("\t\"" + node.Label + "\" [shape=" + shape + ", label=\"" + sb + "\", color=\"1.000 0.000 "
                         + (0.5 - relativeNodeFrequency) + "\"]");
          if (node.OutEdges == null) continue;
          double sum = node.OutEdges.Sum(a => a.Weight);
          foreach (var arc in node.OutEdges) {
            double relativeArcFrequency = arc.Weight / totalArcWeight;
            double normWeight = normalizedWeights[arc];
            sb.Clear();
            sb.Append("\t\"" + node.Label + "\" -> \"" + ((Vertex)arc.Target).Label + "\"[penwidth=" + normWeight);
            sb.Append(", label=\"" + String.Format("{0:0.00%}\\n({1:0.00%})", relativeArcFrequency, arc.Weight / sum));
            sb.Append("\", color=\"1.000 0.000 " + (0.5 - relativeArcFrequency) + "\"]");
            file.WriteLine(sb);
          }
        }

        file.WriteLine("}");
      }
    }
    public static void WriteGml(this IGenericGraph<SymbolNode> graph, string path) {
      var idMap = new Dictionary<string, int>();
      const double minNorm = 1.0, maxNorm = 3.0;
      int idCount = 0;

      var arcs = graph.Nodes.Where(node => node.OutEdges != null).SelectMany(node => node.OutEdges).ToList();
      double totalArcWeight = arcs.Sum(arc => arc.Weight);
      double totalNodeWeight = graph.Nodes.Sum(node => node.Weight);
      var min = arcs.Min(x => x.Weight);
      var max = arcs.Max(x => x.Weight);

      var normalizedWeights = arcs.Select(arc => new KeyValuePair<SymbolArc, double>(arc, (arc.Weight - min) * (maxNorm - minNorm) / (max - min) + minNorm))
                                  .ToDictionary(x => x.Key, x => x.Value);

      var nl = Environment.NewLine;

      using (var file = new StreamWriter(path)) {
        file.WriteLine("graph [" + nl +
                       "\tcomment \"Pruned symbol graph\"" + nl +
                       "\tdirected 1");

        var sb = new StringBuilder();
        foreach (var node in graph.Nodes) {
          if (!idMap.ContainsKey(node.Label))
            idMap.Add(node.Label, idCount++);
          double relativeNodeFrequency = node.Weight / totalNodeWeight;
          sb.Clear();
          sb.Append(node.Label + " (" + String.Format("{0:0.00}", node.AverageArity) + ") " +
                    String.Format("{0:0.00%}", relativeNodeFrequency));
          if (node.Positions != null) {
            double posSum = node.Positions.Sum(x => x.Value);
            //            sb.Append("\\n");
            foreach (var pos in node.Positions)
              sb.Append(String.Format("{0:00%} ", pos.Value / posSum));
          }

          file.WriteLine("\tnode [" + nl +
                         "\t\tid " + idMap[node.Label] + nl +
                         "\t\tlabel \"" + node.Label + "\"" + nl + "\t]");
        }

        foreach (var node in graph.Nodes) {
          if (node.OutEdges == null) continue;
          double sum = node.OutEdges.Sum(a => a.Weight);
          foreach (var arc in node.OutEdges) {
            double relativeArcFrequency = arc.Weight / totalArcWeight;
            double normWeight = normalizedWeights[arc];
            sb.Clear();
            sb.Append("\tedge [" + nl +
                      "\t\tsource " + idMap[node.Label] + nl +
                      "\t\ttarget " + idMap[(arc.Target as SymbolNode).Label] + nl +
                      "\t\tlabel \"" + String.Format("{0:0.00%}) ({1:0.00%})", relativeArcFrequency, arc.Weight / sum) + "\"" + nl + "\t]");
            file.WriteLine(sb);
          }
        }

        file.WriteLine("]");
      }
    }

    public static void WriteDot(this FPGraph graph, string path) {
      var nl = Environment.NewLine;
      const string shape = "rectangle";

      using (var file = new StreamWriter(path)) {
        var sb = new StringBuilder();
        file.WriteLine("digraph {");
        foreach (var level in graph.Levels) {
          foreach (var node in level.Value) {
            file.WriteLine("\t\"" + node.Label + ":" + level.Key + "\" [shape=" + shape + ", label=\"" + node.Label + "\", rank=same]");
          }
        }
        foreach (var level in graph.Levels) {
          foreach (var node in level.Value) {
            if (node.OutEdges == null || node.OutEdges.Count == 0) continue;
            foreach (var arc in node.OutEdges) {
              sb.Clear();
              sb.Append("\t\"" + node.Label + ":" + level.Key + "\" -> \"" + arc.Target.Label + ":" + (level.Key + 1) + "\"");
              file.WriteLine(sb);
            }
          }
        }
        file.WriteLine("}");
      }
    }

    public static string ToHex(this Color color) {
      return "#" + color.R.ToString("X2") + color.G.ToString("X2") + color.B.ToString("X2") + color.A.ToString("X2");
    }
  }
}
